#include "typedef.h"

#include <string.h>
#include <stdlib.h>
#include <search.h>

#include "util.h"
#include "opcode.h"
#include "sort.h"

int		Csort::g_labels_count = 0;
int		Csort::g_labels_max   = 0;
labelT	*Csort::g_labelsP     = 0;	

void	
Csort::clear_sort(void)
{
	g_labels_count = 0;
}

void
Csort::append_sort(int method, int to, int from, pcodeE pcode)
{
	labelT	*labelP;

	if (g_labels_count >= g_labels_max) {
		if (!g_labels_max) {
			g_labels_max = 4096;
			g_labelsP    = (labelT *) xmalloc(g_labels_max * sizeof(labelT));
		} else {
			int		size  = g_labels_max * sizeof(labelT);
			labelT	*newP = (labelT *) xmalloc(size << 1);

			memcpy(newP, g_labelsP, size);
			xfree(g_labelsP);
			g_labelsP = newP;
			g_labels_max <<= 1;
	}	}

	labelP = g_labelsP + g_labels_count++;
	labelP->method = method;
	labelP->to     = to;
	labelP->from   = from;
	labelP->pcode  = pcode;
}

/*
	< 0 P1 less than P2
      0 P1 equals    P2
    > 0 P1 greater   P2
 */

static int
label_compare(const void *P1, const void *P2)
{
	u4T	to1, to2, from1, from2;
	int	diff;

	diff    = ((labelT *) P1)->method - ((labelT *) P2)->method;
	if (diff) {
		return diff;
	}

	to1   = ((labelT *) P1)->to;
	to2   = ((labelT *) P2)->to;
	diff  = (int) to1 - (int) to2;

	if (diff) {
		// to1 occurs before to2 if to1 is smaller
		return (diff);
	}

	from1 = ((labelT *) P1)->from;
	from2 = ((labelT *) P2)->from;
	diff  = (int) from2 - (int) from1;
	if (from1 <= to1) {
		if (from2 > to2) {
			// Labels reached by forward jump occur before labels reached by backward jump
			return -1;
		}
		// Smaller jump  labels occur before longer jump labels
		return (diff);
	}
	if (from2 <= to2) {
		// Labels reached by forward jump occur before labels reached by backward jump
		return 1;
	}
	// Smaller jump labels occur after longer jump labels
	return (diff);
}

void
Csort::sort(void)
{
	qsort(g_labelsP, g_labels_count, sizeof(labelT), label_compare);
}

